# coding=utf-8
from lxml import etree
from requests import Session

import six
from .docket_report import BaseDocketReport
from .docket_utils import normalize_party_types
from .utils import get_docketxml_url, get_pdf_url, is_pdf
from ..lib.judge_parsers import normalize_judge_string
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, convert_date_string, harmonize
from ..lib.utils import clean_court_object, previous_and_next

logger = make_default_logger()

date_regex = r'[—\d\-–/]+'


class InternetArchive(BaseDocketReport):
    """A simple tool for working with the XML and PDFs on the Internet Archive
    in the RECAP collection.
    """

    CACHE_ATTRS = ['metadata', 'parties', 'docket_entries']

    def __init__(self):
        super(InternetArchive, self).__init__()
        # Initialize the empty cache properties.
        self._clear_caches()
        self._metadata = None
        self._parties = None
        self._docket_entries = None

        self.session = Session()
        self.response = None
        self.tree = None
        self.parser = etree.XMLParser(recover=True)
        self.is_valid = True

    def download_pdf(self, pacer_court_id, pacer_case_id, document_number,
                     attachment_number):
        """Download a PDF from the Internet Archive"""
        timeout = (60, 300)
        url = get_pdf_url(pacer_court_id, pacer_case_id, document_number,
                          attachment_number)
        logger.info("GETting PDF at URL: %s")
        r = self.session.get(url, timeout=timeout)
        r.raise_for_status()
        if not is_pdf(r):
            logger.error("Got non-PDF data, but expected a PDF at: %s" % url)
            return None
        else:
            return r

    def query(self, court_id, pacer_case_id):
        """Download a docket XML page from the Internet Archive"""
        timeout = (60, 300)
        url = get_docketxml_url(court_id, pacer_case_id)
        logger.info("GETting docket XML at URL: %s")
        r = self.session.get(url, timeout=timeout)
        self.response = r
        self.parse()

    def parse(self):
        """Parse the item, but clear the cache before you do so."""
        self._clear_caches()
        self.response.raise_for_status()
        self._parse_text(self.response.text)

    def _parse_text(self, text):
        assert isinstance(text, six.text_type), \
            "Input must be unicode, not %s" % type(text)
        self.tree = etree.fromstring(text.encode('utf-8'), self.parser)

    @property
    def metadata(self):
        if self._metadata is not None:
            return self._metadata

        data = {
            u'court_id': self._get_str_from_tree('//court'),
            u'docket_number': self._get_str_from_tree('//docket_num'),
            u'case_name': self._get_case_name(),
            u'date_filed': self.get_datetime_from_tree(
                '//date_case_filed', cast_to_date=True),
            u'date_terminated': self.get_datetime_from_tree(
                '//date_case_terminated', cast_to_date=True),
            u'date_converted': None,
            u'date_discharged': None,
            u'assigned_to_str': self._get_judge('//assigned_to'),
            u'referred_to_str': self._get_judge('//referred_to'),
            u'cause': self._get_str_from_tree('//case_cause'),
            u'nature_of_suit': self._get_str_from_tree('//nature_of_suit'),
            u'jury_demand': self._get_str_from_tree('//jury_demand'),
            u'demand': '',
            u'jurisdiction': self._get_str_from_tree('//jurisdiction'),
        }
        data = clean_court_object(data)
        self._metadata = data
        return data

    @property
    def parties(self):
        """Get the party info from the XML or return it if it's cached."""
        if self._parties is not None:
            return self._parties

        party_nodes = self.tree.xpath('//party_list/party')

        parties = []
        for prev, party_node, nxt in previous_and_next(party_nodes):
            pt = self._xpath_text_0(party_node, './type')
            extra_info = self._xpath_text_0(party_node, './extra_info')
            name = self._xpath_text_0(party_node, './name')
            if not name.strip():
                # Happens in adversary proceedings? See cacb_1669936.xml.
                continue
            party = {
                u'type': normalize_party_types(pt),
                u'name': name,
                u'extra_info': extra_info,
            }

            m = self.date_terminated_regex.search(extra_info)
            if m:
                party['date_terminated'] = convert_date_string(m.group(1))
            else:
                party['date_terminated'] = None

            party[u'attorneys'] = self._get_attorneys(party_node)

            if party not in parties and party != {}:
                # Sometimes there are dups in the docket. Avoid them.
                parties.append(party)

        parties = self._normalize_see_above_attorneys(parties)
        self._parties = parties
        return parties

    def _get_attorneys(self, party_node):
        attorneys = []
        for atty_node in party_node.xpath('./attorney_list/attorney'):
            attorney = {
                u'name': self._xpath_text_0(atty_node, './attorney_name'),
                u'contact': self._xpath_text_0(atty_node, './contact'),
            }
            roles = []
            role_str = self._xpath_text_0(atty_node, './attorney_role')
            for role in role_str.split('\n'):
                role = role.strip()
                if not role:
                    continue
                if not any([role.lower().startswith(u'bar status'),
                            role.lower().startswith(u'designation')]):
                    roles.append(role)
            attorney['roles'] = roles
            attorneys.append(attorney)
        return attorneys

    @property
    def docket_entries(self):
        if self._docket_entries is not None:
            return self._docket_entries

        de_nodes = self.tree.xpath('//document_list/document')
        docket_entries = []
        prev_date_filed = None
        for de_node in de_nodes:
            de = {
                u'document_number': de_node.xpath('./@doc_num')[0],
                u'description': self._xpath_text_0(de_node, './long_desc'),
                u'short_description': self._xpath_text_0(de_node,
                                                         './short_desc'),
                u'pacer_seq_no': self._xpath_text_0(
                    de_node, './pacer_de_seq_num') or None
            }
            attachment_number = de_node.xpath('./@attachment_num')[0]
            if attachment_number != '0':
                de[u'attachment_number'] = attachment_number

            date_filed_str = self._xpath_text_0(de_node, './date_filed')
            if date_filed_str:
                # Got a date. Set it, and save it for the next item.
                try:
                    de[u'date_filed'] = convert_date_string(date_filed_str)
                except ValueError:
                    # Fails for dates like 0000-00-00
                    de[u'date_filed'] = None
                else:
                    prev_date_filed = de[u'date_filed']
            else:
                # No date found.
                if de.get(u'attachment_number'):
                    # If it's an attachment, it probably lacks a date. Get it
                    # from the previously stored item.
                    de[u'date_filed'] = prev_date_filed
                else:
                    # If not an attachment, it's probably an old docket entry,
                    # which sometimes lack dates. Press on.
                    continue

            de[u'pacer_doc_id'] = self._xpath_text_0(de_node,
                                                     './pacer_doc_id') or None

            if not de[u'document_number'].isdigit():
                # Some courts put weird stuff in this column.
                continue
            docket_entries.append(de)

        docket_entries = clean_court_object(docket_entries)
        self._docket_entries = docket_entries
        return docket_entries

    def _get_judge(self, path):
        judge_str = self._get_str_from_tree(path)
        if judge_str is not None:
            return normalize_judge_string(judge_str)[0]

    def _get_case_name(self):
        case_name = self._get_str_from_tree('//case_name')
        case_name = clean_string(harmonize(case_name))
        if not case_name:
            return u"Unknown Case Title"
        return case_name




"""Scraper for United States District Court for the District of Columbia
CourtID: dcd
Court Short Name: D.D.C.
Author: V. David Zvenyach
Date created: 2014-02-27
Substantially Revised: Brian W. Carver, 2014-03-28
"""

import time
from datetime import date
from lxml import html
import re

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://ecf.dcd.uscourts.gov/cgi-bin/Opinions.pl?' + str(date.today().year)

    def _get_download_urls(self):
        # There are often multiple documents and hence urls for each case.
        # This requires us to pad every other metadata field to match the
        # number of urls we find here.
        path = '//table[2]//tr[position()>0]/td[3]/a/@href'
        return [url for url in self.html.xpath(path)]

    def _get_case_names(self):
        casenames = []
        rowpath = '//table[2]//tr[position()>0]'
        cnpath = './td[2]//text()[preceding-sibling::br]'
        urlpath = './td[3]/a/@href'
        for row in self.html.xpath(rowpath):
            case_list = row.xpath(cnpath)
            for rough_case_name in case_list:
                case_name = titlecase(rough_case_name.lower())
                # Determine the number of urls in each row and pad the case
                # name list sufficiently
                count = len(row.xpath(urlpath))
                casenames.extend([case_name] * count)
        return casenames

    def _get_case_dates(self):
        dates = []
        rowpath = '//table[2]//tr[position()>0]'
        datepath = './td[1]/text()'
        urlpath = './td[3]/a/@href'
        for row in self.html.xpath(rowpath):
            date_string = row.xpath(datepath)
            for d in date_string:
                date_object = date.fromtimestamp(
                      time.mktime(time.strptime(d, '%m/%d/%Y')))
                # Determine the number of urls in each row and pad the date
                # list sufficiently
                count = len(row.xpath(urlpath))
                dates.extend([date_object] * count)
        return dates

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_docket_numbers(self):
        docket_numbers = []
        rowpath = '//table[2]//tr[position()>0]'
        dktpath = './td[2]//text()[following-sibling::br]'
        urlpath = './td[3]/a/@href'
        for row in self.html.xpath(rowpath):
            docket_number = row.xpath(dktpath)
            # Determine the number of urls in each row and pad the docket
            # numbers list sufficiently
            count = len(row.xpath(urlpath))
            docket_numbers.extend(docket_number * count)
        return docket_numbers

    def _get_docket_document_numbers(self):
        document_numbers = []
        regex = re.compile('(\?)(\d+)([a-z]+)(\d+)(-)(.*)')
        for url in self.html.xpath('//table[2]//tr[position()>0]/td[3]/a/@href'):
            # Because we are acting directly on the entire url list, no padding
            # of the docket number field is required.
            doc_no = regex.search(url)
            # In 2012 (and perhaps elsewhere) they have a few weird urls.
            if re.search(regex, url) is not None:
                document_numbers.append(doc_no.group(6))
            else:
                document_numbers.append(url)
        return document_numbers

    def _get_judges(self):
        judges = []
        rowpath = '//table[2]//tr[position()>0]'
        urlpath = './td[3]/a/@href'
        judgepath = './td[3]'
        for row in self.html.xpath(rowpath):
            for judge_element in row.xpath(judgepath):
                judge_string = html.tostring(judge_element, method='text', encoding='unicode')
                judge = re.search('(by\s)(.*)', judge_string, re.MULTILINE).group(2)
                # Determine the number of urls in each row and pad the judges
                # list sufficiently
                count = len(row.xpath(urlpath))
                judges.extend([judge] * count)
        return judges

    def _get_nature_of_suit(self):
        nos = []
        for url in self.html.xpath('//table[2]//tr[position()>0]/td[3]/a/@href'):
            # Because we are acting directly on the entire url list, no padding
            # of the nature of suit field is required.
            regex = '(\?)(\d+)([a-z]+)(\d+)(\-)(.*)'
            url_str = re.search(r'(\?)(\d+)([a-z]+)(\d+)(-)(.*)',url)
            # In 2012 (and perhaps elsewhere) they have a few weird urls.
            if re.search(regex, url) is not None:
                nature_code = url_str.group(3)
                if nature_code == 'cv':
                    nos.append("Civil")
                elif nature_code == 'cr':
                    nos.append("Criminal")
                # This is a tough call. Magistrate Cases are typically also
                # Criminal or Civil cases, and their docket_number field will
                # reflect this, but they do classify these separately under
                # these 'mj' and 'mc' codes and the first page of these
                #  documents will often refer to them as 'Magistrate Case
                # ####-####' so, we will too.
                elif nature_code == 'mj' or 'mc':
                    nos.append("Magistrate Case")
                else:
                    nos.append("Unknown")
            else:
                nos.append("Unknown")
        return nos



# -*- coding: utf-8 -*-
import glob
import json
from collections import defaultdict
from pprint import pprint

import re
import requests
from juriscraper.AbstractSite import AbstractSite
from datetime import date

from lxml import etree
from requests.exceptions import MissingSchema


def get_tree(url):
    try:
        response = requests.get(url, stream=True)
        response.raw.decode_content = True
        return etree.parse(response.raw)
    except MissingSchema:
        return etree.parse(url)


def xpath(tree, query):
    return tree.xpath(
        query,
        namespaces={
            'm': 'http://www.loc.gov/mods/v3',
            's': 'http://www.sitemaps.org/schemas/sitemap/0.9',
            'xlink': 'http://www.w3.org/1999/xlink',
        }
    )


class FDSysModsContent(object):

    def __init__(self, url):

        self._all_attrs = [
            'download_url',
            'fdsys_id',
            'court_id',
            'docket_number',
            'court_location',
            'parties',
            'case_name',
            'documents'
        ]
        self.tree = None
        mods_url = self._get_mods_file_url(url)
        self.parse(mods_url)

    def get_content(self):
        """Using i, convert a single item into a dict. This is effectively a
        different view of the data.
        """
        item = {}
        for attr_name in self._all_attrs:
            item[attr_name] = getattr(self, attr_name)
        return item

    def parse(self, url):
        self.tree = get_tree(url)

        for attr in self._all_attrs:
            self.__setattr__(attr, getattr(self, '_get_%s' % attr)())

    def _get_download_url(self):
        return ''.join(xpath(self.tree, "(//m:identifier[@type='uri'])[1]/text()")).strip()

    def _get_fdsys_id(self):
        return ''.join(xpath(self.tree, "(//m:accessId/text())[1]"))

    def _get_docket_number(self):
        return ''.join(xpath(self.tree, "(//m:caseNumber/text())[1]"))

    def _get_court_id(self):
        return ''.join(xpath(self.tree, "(//m:courtCode/text())[1]")).lower()

    def _get_court_location(self):
        return ''.join(xpath(self.tree, "(//m:caseOffice/text())[1]"))

    def _get_case_name(self):
        return ''.join(xpath(self.tree, "(//m:titleInfo/m:title/text())[1]"))

    def _get_parties(self):
        """Extract the parties from the XML into a nice object."""
        party_nodes = xpath(self.tree, "(//m:extension[m:party])[1]//m:party")

        return map(self._get_party, party_nodes)

    @staticmethod
    def _get_party(party_node):
        return {
            'name_first': ''.join(xpath(party_node, './@firstName')),
            'name_last': ''.join(xpath(party_node, './@lastName')),
            'name_middle': ''.join(xpath(party_node, './@middleName')),
            'name_suffix': ''.join(xpath(party_node, './@generation')),
            'role': ''.join(xpath(party_node, './@role')),
        }

    def _get_documents(self):
        """Get the documents from the XML into a nice object."""
        document_nodes = xpath(self.tree, "//m:mods/m:relatedItem")

        return map(self._get_document, document_nodes)

    def _get_document(self, document_node):
        desription = ' '.join(''.join(
                            xpath(document_node, './/m:subTitle/text()')
                    ).split()
                              )
        return {
            'download_url': ''.join(xpath(document_node, './m:relatedItem/@xlink:href')).strip(),
            'description': desription,
            'date_filed': ''.join(xpath(document_node, './m:originInfo/m:dateIssued/text()')),
            # 'type': self._get_document_type(desription),
            'number': ''.join(xpath(document_node, './/m:partNumber/text()')),
        }

    @staticmethod
    def _get_mods_file_url(url):
        """replaces content-detail.html with mods.xml"""
        return url.replace('content-detail.html', 'mods.xml')

    # def _get_document_type(self, description):
    #     # get the first 5 words
    #     return ''


class FDSysSite(AbstractSite):
    """Contains generic methods for scraping fdsys. Should be extended by all
    scrapers for fdsys.

    Should not contain lists that can't be sorted by the _date_sort function.


    """

    def __init__(self, *args, **kwargs):
        super(FDSysSite, self).__init__(*args, **kwargs)
        current_year = date.today().year
        self.base_url = "https://www.gpo.gov/smap/fdsys/sitemap_{year}/{year}_USCOURTS_sitemap.xml"
        self.url = self.base_url.format(year=current_year)
        self.back_scrape_iterable = range(1982, current_year + 1)

    def __iter__(self):
        for i, url in enumerate(xpath(self.html, "//s:loc/text()")):
            self.save_mods_file(url)
            mods_file = FDSysModsContent(url)
            yield mods_file.get_content()

    def __getitem__(self, i):
        mods_file = FDSysModsContent(xpath(self.html, "//s:loc/text()")[i])
        return mods_file.get_content()

    def __len__(self):
        return len(xpath(self.html, "//s:loc/text()"))

    def save_mods_file(self, url):
        mods_url = FDSysModsContent._get_mods_file_url(url)
        name = '-'.join(mods_url.split("/")[-2].split('-')[1:])
        with open("./examples/2006/{}.xml".format(name), 'w') as handle:
            response = requests.get(mods_url, stream=True)
            for block in response.iter_content(1024):
                handle.write(block)

    def _download(self, request_dict={}):
        """
        it actually builds an XML tree
        """
        return get_tree(self.url)

    def _download_backwards(self, year):
        self.url = self.base_url.format(year=year)
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200

    def _check_sanity(self):
        pass

    def parse(self):
        if self.status is None:
            # Run the downloader if it hasn't been run already
            self.html = self._download()
        return self


def get_court_locations_list():
    """
    parses the examples directories and gets the court ids and the court locations
    """
    court_locations_list = defaultdict(set)
    # parse all the example files
    for f in glob.glob('./examples/*/*.xml'):
        fm = FDSysModsContent(f)
        print(f, fm.court_id, fm.court_location)
        court_locations_list[fm.court_id].add(fm.court_location)

    # change set to list
    cl = {}
    for k, v in court_locations_list.items():
        cl[k] = list(v)

    # save as json
    with open('court_locations.json', 'w') as j:
        json.dump(cl, j)


def get_the_first_5_words():
    l = defaultdict(list)
    p = defaultdict(list)
    word_counter = defaultdict(int)
    for f in glob.glob('./examples/*/*.xml'):
        fm = FDSysModsContent(f)
        print(f, fm.court_id, fm.court_location)
        for document in fm.documents:
            try:
                words_to_use = document['description'].split()[:8]
            except IndexError:
                words_to_use = document['description'].split()
            ws = []
            for word in words_to_use:
                try:
                    w = re.findall('[a-zA-Z]+', word)[0]
                except IndexError:
                    w = None

                if w:
                    word_counter[w] += 1
                ws.append(w)
            l[f.__repr__()].append(ws)
            p[f.__repr__()].append(" ".join(words_to_use))

    with open('first_8_words_string.json', 'w') as pc:
        json.dump(p, pc)

    with open('first_five_words.json', 'w') as j:
        json.dump(l, j)

    with open('first_five_words_counter.json', 'w') as c:
        json.dump(word_counter, c)


if __name__ == '__main__':
    # get_court_locations_list()
    get_the_first_5_words()
    # for f in glob.glob('./examples/*/*.xml'):
    #     fm = FDSysModsContent(f)
    #
    #     pprint(fm.get_content())
    # f = FDSysSite()
    # f.url = "./sitemaps_examples/2006_USCOURTS_sitemap.xml"
    # f.parse()
    # for i in f:
    #     print(i)
    # #     pprint(i)0



import codecs
import os
import sys
import unittest

try:  # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements

from setuptools import setup, find_packages
from setuptools.command.install import install

VERSION = "1.26.15"
AUTHOR = "Free Law Project"
EMAIL = "info@free.law"
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


class TestNetwork(install):
    """Run network test only"""
    description = 'run isolated tests that hit the network'

    def run(self):
        loader = unittest.defaultTestLoader
        runner = unittest.TextTestRunner(verbosity=2)
        tests = loader.discover('./tests/network')
        runner.run(tests)


class VerifyVersion(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag is None:
            sys.exit("The 'verify' option is only available in tagged CircleCI container")

        if tag != VERSION:
            message = "Git tag: {0} does not match the version of this app: {1}"
            sys.exit(message.format(tag, VERSION))


requirements = [
    str(r.req) for r in
    parse_requirements('requirements.txt', session=False)
]

setup(
    name="juriscraper",
    description="An API to scrape American court websites for metadata.",
    license="BSD",
    url="https://github.com/freelawproject/juriscraper",
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    maintainer=AUTHOR,
    maintainer_email=EMAIL,
    keywords=["scraping", "legal", "pacer"],
    long_description=read("README.rst"),
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=requirements,
    tests_require=['jsondate', 'mock', 'vcrpy'],
    include_package_data=True,
    test_suite='tests.test_local',
    cmdclass={
        'verify': VerifyVersion,
        'testnetwork': TestNetwork,
    }
)



import re

import requests
from requests.packages.urllib3 import exceptions

from ..lib.exceptions import PacerLoginException
from ..lib.html_utils import get_html_parsed_text, get_xml_parsed_text
from ..lib.log_tools import make_default_logger
from ..pacer.utils import is_pdf

logger = make_default_logger()

requests.packages.urllib3.disable_warnings(exceptions.InsecureRequestWarning)


def check_if_logged_in_page(text):
    """Is this a valid HTML page from PACER?

    Check if the html in 'text' is from a valid PACER page or valid PACER XML
    document, or if it's from a page telling you to log in or informing you
    that you're not logged in.
    :param text: The HTML or XML of the page to test
    :return boolean: True if logged in, False if not.
    """
    valid_case_number_query = '<case number=' in text or \
                              "<request number=" in text
    no_results_case_number_query = re.search('<message.*Cannot find', text)
    sealed_case_query = re.search('<message.*Case Under Seal', text)
    if any([valid_case_number_query, no_results_case_number_query,
            sealed_case_query]):
        not_logged_in = re.search('text.*Not logged in', text)
        if not_logged_in:
            # An unauthenticated PossibleCaseNumberApi XML result. Simply
            # continue onwards. The complete result looks like:
            # <request number='1501084'>
            #   <message text='Not logged in.  Please refresh this page.'/>
            # </request>
            # An authenticated PossibleCaseNumberApi XML result.
            return False
        else:
            return True

    # Detect if we are logged in. If so, no need to do so. If not, we login
    # again below.
    found_district_logout_link = '/cgi-bin/login.pl?logout' in text
    found_appellate_logout_link = 'InvalidUserLogin.jsp' in text
    if any([found_district_logout_link, found_appellate_logout_link]):
        # A normal HTML page we're logged into.
        return True

    return False


class PacerSession(requests.Session):
    """
    Extension of requests.Session to handle PACER oddities making it easier
    for folks to just POST data to PACER endpoints/apis.

    Also includes utilities for logging into PACER and re-logging in when
    sessions expire.
    """
    LOGIN_URL = 'https://pacer.login.uscourts.gov/csologin/login.jsf'
    INDEX = 'https://ecf.ilnd.uscourts.gov/cgi-bin/ShowIndex.pl'

    def __init__(self, cookies=None, username=None, password=None):
        """
        Instantiate a new PACER HTTP Session with some Juriscraper defaults
        :param pacer_token: a PACER_SESSION token value
        """
        super(PacerSession, self).__init__()
        self.headers['User-Agent'] = 'Juriscraper'
        self.headers['Referer'] = 'https://external'  # For CVE-001-FLP.
        self.verify = False

        if cookies:
            self.cookies = cookies

        self.username = username
        self.password = password

    def get(self, url, auto_login=True, **kwargs):
        """Overrides request.Session.get with session retry logic.

        :param url: url string to GET
        :param auto_login: Whether the auto-login procedure should happen.
        :return: requests.Response
        """
        kwargs.setdefault('timeout', 300)

        r = super(PacerSession, self).get(url, **kwargs)

        if 'This user has no access privileges defined.' in r.text:
            # This is a strange error that we began seeing in CM/ECF 6.3.1 at
            # ILND. You can currently reproduce it by logging in on the central
            # login page, selecting "Court Links" as your destination, and then
            # loading: https://ecf.ilnd.uscourts.gov/cgi-bin/WrtOpRpt.pl
            # The solution when this error shows up is to simply re-run the get
            # request, so that's what we do here. PACER needs some frustrating
            # and inelegant hacks sometimes.
            r = super(PacerSession, self).get(url, **kwargs)
        if auto_login:
            updated = self._login_again(r)
            if updated:
                # Re-do the request with the new session.
                return super(PacerSession, self).get(url, **kwargs)
        return r

    def post(self, url, data=None, json=None, auto_login=True, **kwargs):
        """
        Overrides requests.Session.post with PACER-specific fun.

        Will automatically convert data dict into proper multi-part form data
        and pass to the files parameter instead.

        Will set a timeout of 300 if not provided.

        All other uses or parameters will pass through untouched
        :param url: url string to post to
        :param data: post data
        :param json: json object to post
        :param auto_login: Whether the auto-login procedure should happen.
        :param kwargs: assorted keyword arguments
        :return: requests.Response
        """
        kwargs.setdefault('timeout', 300)

        if data:
            pacer_data = self._prepare_multipart_form_data(data)
            kwargs.update({'files': pacer_data})
        else:
            kwargs.update({'data': data, 'json': json})

        r = super(PacerSession, self).post(url, **kwargs)
        if auto_login:
            updated = self._login_again(r)
            if updated:
                # Re-do the request with the new session.
                return super(PacerSession, self).post(url, **kwargs)
        return r

    def head(self, url, **kwargs):
        """
        Overrides request.Session.head with a default timeout parameter.

        :param url: url string upon which to do a HEAD request
        :param kwargs: assorted keyword arguments
        :return: requests.Response
        """
        kwargs.setdefault('timeout', 300)
        return super(PacerSession, self).head(url, **kwargs)

    @staticmethod
    def _prepare_multipart_form_data(data):
        """
        Transforms a data dictionary into the multi-part form data that PACER
        expects as the POST body
        :param data: dict of data to transform
        :return: dict with values wrapped into tuples like:(None, <value>)
        """
        output = dict()
        for key in data:
            output[key] = (None, data[key])
        return output

    @staticmethod
    def _get_view_state(r):
        """Get the viewState parameter of the form

        This is an annoying thing we have to do. The login flow has three
        requests that you make and each requires the view state from the one
        prior. Thus, we capture that viewState each time and submit it during
        each of the next submissions.

        The HTML takes the form of:

        <input type="hidden" name="javax.faces.ViewState"
               id="j_id1:javax.faces.ViewState:0"
               value="some-long-value-here">

        :param r: A request.Response object
        :return The value of the "value" attribute of the ViewState input
        element.
        """
        tree = get_html_parsed_text(r.content)
        xpath = '//form[@id="loginForm"]//input[' \
                '    @name="javax.faces.ViewState"' \
                ']/@value'
        return tree.xpath(xpath)[0]

    @staticmethod
    def _get_xml_view_state(r):
        """Same idea as above, but sometimes PACER returns XML so we parse
        that instead of the HTML.

        Here's a sample of the XML:

        <partial-response id="j_id1">
          <changes>
            <update id="regmsg:bpmConfirm">
              <![CDATA[<button id="regmsg:bpmConfirm" name="regmsg:bpmConfirm" class="ui-button ui-widget ui-state-default ui-corner-all ui-button-text-only" onclick="PrimeFaces.ab({s:&quot;regmsg:bpmConfirm&quot;,pa:[{name:&quot;dialogName&quot;,value:&quot;redactionDlg&quot;}]});return false;" style="margin-right: 20px;" type="submit"><span class="ui-button-text ui-c">Continue</span></button><script id="regmsg:bpmConfirm_s" type="text/javascript">PrimeFaces.cw("CommandButton","widget_regmsg_bpmConfirm",{id:"regmsg:bpmConfirm"});</script>]]>
            </update>
            <update id="loginForm:pclLoginMessages">
              <![CDATA[<div id="loginForm:pclLoginMessages" class="ui-messages ui-widget" style="font-size: 0.85em;" aria-live="polite"></div>]]>
            </update>
            <update id="j_id1:javax.faces.ViewState:0"><![CDATA]]>
            </update>
          </changes>
        </partial-response>
        """
        tree = get_xml_parsed_text(r.content)
        xpath = "//update[@id='j_id1:javax.faces.ViewState:0']/text()"
        return tree.xpath(xpath)[0]

    def login(self, url=None):
        """Attempt to log into the PACER site.

        Logging into PACER has two flows. If you have filing permission in any
        court, you wind up making four POST request which are tied together by
        a JSESSIONID cookie value and ViewState hidden form inputs.

        If you do *not* have filing permissions, only the first request below
        is needed. The trick to determine what's needed is to watch for a 302
        response status or the cookies to be set properly. If you get one or
        the other, that indicates that you're logged in or that you're being
        redirected to the correct court/webpage.

        Here are the requests that are needed. First, just GET the login page
        and parse out the ViewState form value. Then, submit your user/pass:

            curl 'https://pacer.login.uscourts.gov/csologin/login.jsf' \
              -H 'Content-Type: application/x-www-form-urlencoded' \
              --data 'login=login&login%3AloginName=mlissner.flp&login%3Apassword=QKAXmos0DyAtpX%26U30O7Vqt%401NNwa3z%5E&login%3AclientCode=&login%3AfbtnLogin=&javax.faces.ViewState=stateless' \
              --verbose > /tmp/curl-out.html

        If this is *not* a filing account, you should receive a 302 response
        and the proper cookies at this point. You're logged in.

        If this *is* a filing account, the third request happens when you
        click the *box* (not the button) saying you'll agree to the redaction
        rules. Note that the JSESSIONID cookie in this request and the next one
        is set by the previous request and needs to be carried through. If you
        receive a new JSESSIONID cookie in response to your second request,
        something has gone wrong. We also grab the ViewState value from the
        previous request:

            curl 'https://pacer.login.uscourts.gov/csologin/login.jsf' \
              -H 'Content-Type: application/x-www-form-urlencoded; charset=UTF-8' \
              -H 'Cookie: JSESSIONID=C73BDC1E4CA8C5BDEE13EB2F4CB75E06' \
              --data 'javax.faces.partial.ajax=true&javax.faces.source=regmsg%3AchkRedact&javax.faces.partial.execute=regmsg%3AchkRedact&javax.faces.partial.render=regmsg%3AbpmConfirm&javax.faces.behavior.event=valueChange&javax.faces.partial.event=change&regmsg=regmsg&regmsg%3AchkRedact_input=on&javax.faces.ViewState=stateless' \
              --verbose > /tmp/curl-out.html

        The fourth request happens when you submit the form saying you promise
        to redact:

            curl 'https://pacer.login.uscourts.gov/csologin/login.jsf' \
              -H 'Content-Type: application/x-www-form-urlencoded' \
              -H 'Cookie: JSESSIONID=C73BDC1E4CA8C5BDEE13EB2F4CB75E06' \
              --data 'regmsg=regmsg&regmsg%3AchkRedact_input=on&regmsg%3AbpmConfirm=&javax.faces.ViewState=stateless'\
              --verbose > /tmp/curl-out.html

        If you get a 302 response and the proper cookies at this point, that
        means you're logged in.
        """
        logger.info(u'Attempting PACER site login')

        # Clear any remaining cookies. This is important because sometimes we
        # want to login before an old session has entirely died. One example of
        # when we do that is when we get the page saying that "This page will
        # expire in...[so many minutes]." When we see that we just log in
        # fresh and try again.
        self.cookies.clear()
        if url is None:
            url = self.LOGIN_URL

        # Load the page in order to get the ViewState value from the HTML
        load_page_r = self.get(
            url,
            headers={'User-Agent': 'Juriscraper'},
            auto_login=False,
            verify=False,
            timeout=60,
        )

        login_post_r = self.post(
            url,
            headers={'User-Agent': 'Juriscraper'},
            verify=False,
            timeout=60,
            auto_login=False,
            data={
                'javax.faces.partial.ajax': 'true',
                'javax.faces.partial.execute': '@all',
                'javax.faces.source': 'loginForm:fbtnLogin',
                'javax.faces.partial.render':
                    'pscLoginPanel+loginForm+redactionConfirmation+popupMsgId',
                'javax.faces.ViewState': self._get_view_state(load_page_r),

                'loginForm:courtId_input': 'E_ALMDC',
                'loginForm:courtId_focus': '',
                'loginForm:fbtnLogin': 'loginForm:fbtnLogin',
                'loginForm:loginName': self.username,
                'loginForm:password': self.password,
                'loginForm:clientCode': '',
                'loginForm': 'loginForm',

            },
        )
        if u'Invalid username or password' in login_post_r.text:
            raise PacerLoginException("Invalid username/password")
        if u'Username must be at least 6 characters' in login_post_r.text:
            raise PacerLoginException("Username must be at least six "
                                      "characters")
        if u'Password must be at least 8 characters' in login_post_r.text:
            raise PacerLoginException("Password must be at least eight "
                                      "characters")
        if u'timeout error' in login_post_r.text:
            raise PacerLoginException("Timeout")

        if not self.cookies.get('PacerSession'):
            logger.info("Did not get cookies from first log in POSTs. "
                        "Assuming this is a filing user and doing two more "
                        "POSTs.")
            reg_msg_r = self.post(
                url,
                headers={'User-Agent': 'Juriscraper'},
                verify=False,
                timeout=60,
                auto_login=False,
                data={
                    'javax.faces.partial.ajax': 'true',
                    'javax.faces.source': 'regmsg:chkRedact',
                    'javax.faces.partial.execute': 'regmsg:chkRedact',
                    'javax.faces.partial.render': 'regmsg:bpmConfirm',
                    'javax.faces.partial.event': 'change',
                    'javax.faces.behavior.event': 'valueChange',
                    'javax.faces.ViewState': self._get_xml_view_state(
                        login_post_r),

                    'regmsg': 'regmsg',
                    'regmsg:chkRedact_input': 'on',
                },
            )
            # The box is now checked. Submit the form to say so.
            self.post(
                url,
                headers={'User-Agent': 'Juriscraper'},
                verify=False,
                timeout=60,
                auto_login=False,
                data={
                    'javax.faces.partial.ajax': 'true',
                    'javax.faces.source': 'regmsg:bpmConfirm',
                    'javax.faces.partial.execute': '@all',
                    'javax.faces.ViewState': self._get_xml_view_state(
                        reg_msg_r),

                    'regmsg': 'regmsg',
                    'regmsg:chkRedact_input': 'on',
                    'regmsg:bpmConfirm': 'regmsg:bpmConfirm',

                    'dialogName': 'redactionDlg',
                },
            )

        if not self.cookies.get('PacerSession') and \
                not self.cookies.get('NextGenCSO'):
            raise PacerLoginException(
                'Did not get PacerSession and NextGenCSO '
                'cookies when attempting PACER login.'
            )

        self.get(self.INDEX, auto_login=False)
        logger.info(u'New PACER session established.')

    def _login_again(self, r):
        """Log into PACER if the session has credentials and the session has
        expired.

        :param r: A response object to inspect for login errors.
        :returns: A boolean indicating whether a new session needed to be
        created.
        :raises: PacerLoginException, if unable to create a new session.
        """
        if is_pdf(r):
            return False

        logged_in = check_if_logged_in_page(r.text)
        if logged_in:
            return False

        if self.username and self.password:
            logger.info(u"Invalid/expired PACER session. Establishing new "
                        u"session.")
            self.login()
            return True
        else:
            msg = (u"Invalid/expired PACER session and do not have "
                   u"credentials for re-login.")
            logger.error(msg)
            raise PacerLoginException(msg)



from juriscraper.opinions.united_states.federal_special import cit

import time
from datetime import date
from lxml import html


class Site(cit.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2010.html'
        self.court_id = self.__module__

    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath('//table[3]/tr/td[3][../td/a]'):
            date_string = html.tostring(e, method='text', encoding='unicode').strip()
            if date_string == "06/25//2010":
                # Special case.
                date_string = "06/25/2010"
            case_dates.append(date.fromtimestamp(
                         time.mktime(time.strptime(date_string, '%m/%d/%Y'))))
        return case_dates

    def _get_download_urls(self):
        return [t for t in self.html.xpath('//table[3]//tr/td[1]/a/@href')]

    def _get_neutral_citations(self):
        return [t for t in self.html.xpath('//table[3]//tr/td[1]/a/text()')]

    def _get_case_names(self):
        # Exclude confidential rows
        case_names = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[2]'):
            s = html.tostring(e, method='text', encoding='unicode')
            if "confidential" in s:
                continue
            elif 'Errata' in s:
                index = s.find('Errata')
            elif 'Public version' in s:
                index = s.find('Public version')
            else:
                index = -1
            case_names.append(s[:index] if index > -1 else s)
        return case_names

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[2][../td/a]'):
            s = html.tostring(e, method='text', encoding='unicode').lower().strip()
            if "errata" in s:
                statuses.append('Errata')
            else:
                statuses.append('Published')
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[4][../td/a]'):
            docket_numbers.append(html.tostring(
                                e, method='text', encoding='unicode').strip())
        return docket_numbers

    def _get_judges(self):
        judges = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[5][../td/a]'):
            s = html.tostring (e, method='text', encoding='unicode')
            judges.append(s)
        return judges

    def _get_nature_of_suit(self):
        return [t for t in self.html.xpath('//table[3]//tr/td[6][../td/a]/text()')]




"""
Scraper for Massachusetts Appeals Court
CourtID: massapp
Court Short Name: MS
Author: Andrei Chelaru
Court Contact: SJCReporter@sjc.state.ma.us (617) 557-1030
Reviewer: mlr
Date: 2014-07-12
"""

from juriscraper.opinions.united_states.state import mass


class Site(mass.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_identifier = 'AC'
        self.set_local_variables()



from juriscraper.OpinionSite import OpinionSite
import time
from datetime import date
from datetime import timedelta
from juriscraper.lib.string_utils import clean_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        begin_date = date.strftime(date.today() - timedelta(15), '%m/%d/%Y')
        end_date = date.strftime(date.today(), '%m/%d/%Y')
        self.url = ('http://wicourts.gov/supreme/scopin.jsp?'
                   'begin_date=%s&end_date=%s&SortBy=date') % (begin_date,
                                                               end_date)
        self.court_id = self.__module__

    def _get_case_names(self):
        return [t for t in self.html.xpath('//form/table[1]//tr/td[3]/text()')]

    def _get_download_urls(self):
        return [href for href in
                self.html.xpath('//form/table[1]/tr//td[4]/a[contains(.//text(), "PDF")]/@href')]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath('//form/table[1]/tr//td[1]/text()'):
            dates.append(date.fromtimestamp(time.mktime(time.strptime(clean_string(date_string), '%b %d, %Y'))))
        return dates

    def _get_docket_numbers(self):
        return [e for e in self.html.xpath('//form/table[1]/tr//td[2]/text()')]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



"""
History:
 - 2014-08-05: Adapted scraper to have year-based URLs.
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.courts.oregon.gov/publications/sc/Pages/default.aspx'
        self.cases = []
        self.status = 'Published'

    def _process_html(self):
        for header in self.html.xpath('//h4//a/parent::h4'):
            date_string = header.text_content().strip()
            if not date_string:
                continue
            ul = header.xpath('./following-sibling::ul')[0]
            for item in ul.xpath('.//li'):
                # Ensure two links are present (skip Petitions for Review rows)
                # see or_example_2.html
                anchors = item.xpath('.//a')
                if not (len(anchors) > 1):
                    continue
                text = item.text_content().strip()
                url = anchors[0].xpath('./@href')[0]
                docket = anchors[1].text_content().strip()
                name = text.split(')', 1)[-1]
                self.cases.append({
                    'date': date_string,
                    'name': name,
                    'docket': docket,
                    'url': url,
                })



"""
Author: Deb Linton
Date created: 2014-02-14
Scraper for the Court of Appeals of Arizona, Division 1
CourtID: arizctapp
Court Short Name: Ariz. Ct. App.
"""

from juriscraper.opinions.united_states.state import ariz


class Site(ariz.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.azcourts.gov/opinions/SearchOpinionsMemoDecs.aspx?court=998'



from lxml import etree

from .reports import BaseReport
from .utils import get_pacer_doc_id_from_doc1_url
from ..lib.diff_tools import get_closest_match_index
from ..lib.exceptions import ParsingException
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import force_unicode

logger = make_default_logger()


class PossibleCaseNumberApi(BaseReport):
    """Tools for gathering data from the "Possible case numbers" hidden API.

    This API takes a docket number and a court as input, and returns an XML
    object with the pacer_case_id, title, normalized docket number, and a few
    other things.

    This API is normally used via AJAX in the front end of PACER when you put a
    docket number into the docket report search box.
    """
    PATH = 'cgi-bin/possible_case_numbers.pl'

    def query(self, docket_number):
        """Query the "possible case numbers" endpoint and return the results.

        :param docket_number: A string representing a docket number
        :return: a request response object
        """
        assert self.session is not None, \
            "session attribute of PossibleCaseNUmberApi cannot be None."
        url = "%s?%s" % (self.url, docket_number.lower())
        logger.info(u'Querying the possible case number endpoint at URL: %s' %
                    url)
        self.response = self.session.get(url)
        self.parse()

    def _parse_text(self, text):
        """Parse the text of the response object and put it in self.tree.

        Use an XML parser instead of the normal HTML one.
        """
        self.tree = etree.fromstring(text)

    def data(self, case_name=None, office_number=None,
             docket_number_letters=None):
        """Get data back from this query for the matching case.

        :param case_name: This endpoint can return multiple cases for a given
        query. If a case name is provided, and more than one case is returned,
        we will try to identify the most similar case.
        :param office_number: The office number where the item was litigated.
        Usually, this will show up as part of the extended docket number, before
        the colon. For example, in this instance the office number is 2:
        2:16-cr-01152-JZB. Sometimes we know the office number, for example, if
        we're using the IDB. Note that office "numbers" can sometimes be
        a letter. The numbering appears to be 1-9, then A-Z.
        :param docket_number_letters: Sometimes you'll know the letters that
        should be in the returned docket number. For example, they might be 'cr'
        or 'cv'. If you do, you can add this parameter to help filter the
        possible results based on the docket numbers returned.
        For example, if you set this to "cv", and you get back two results:
          2:16-cr-01152-JZB
          2:16-cv-01152-JZB
        You'll know that it's the second result because cv means civil. Again
        this is something you might know if you're using the IDB. These are
        some known values:
         - cv - Civil
         - cr - Criminal
         - ap - Appellate (bankruptcy?)
         - mdl - Multi-district litigation
         - bk - Bankruptcy
        There are others too, feel free to add them if you know them.
        """
        case_count = self.tree.xpath('count(//case)')
        nodes = self.tree.xpath('//case')
        if case_count == 0:
            try:
                msg = self.tree.xpath('//message/@text')[0]
            except IndexError:
                pass
            else:
                if "Cannot find case" in msg or "Case Under Seal" in msg:
                    logger.info("Cannot find case.")
                    return None
            raise ParsingException("Unknown XML content in "
                                   "PossibleCaseNumberApi result.")
        elif case_count == 1:
            # Only one node, all set to go.
            pass
        elif case_count > 1:
            # Filter by things we know
            if office_number is not None:
                # Remove items by docket number, if they lead with something
                # other than the correct office number.
                def correct_office_number(node):
                    try:
                        number = node.xpath('./@number')[0].split(':')[0]
                    except IndexError:
                        # Don't filter...something went wrong.
                        return True
                    return number == office_number

                nodes = list(filter(correct_office_number, nodes))

            if len(nodes) > 1 and docket_number_letters is not None:
                # Remove items by docket number, if they have cv or cr.
                f = lambda node: docket_number_letters in node.xpath('./@number')[0]
                nodes = list(filter(f, nodes))

            if len(nodes) > 1:
                # If we only have sequential defendant attributes, pick the
                # lowest one. The lowest one tends to be the main case, and the
                # others are additional dockets for specific defendants.
                try:
                    attribute = '@defendant'
                    ids = sorted([int(x.xpath('./%s' % attribute)[0])
                                  for x in nodes])
                except IndexError:
                    # No defendant attribute one of the nodes. Try by
                    # pacer_case_id instead. We used to do this by default,
                    # instead of defendant number, but it is not always
                    # sequential.
                    attribute = '@id'
                    ids = sorted([int(x.xpath('./%s' % attribute)[0])
                                  for x in nodes])

                missing_ids = set(range(ids[0], ids[-1] + 1)).difference(ids)
                if len(missing_ids) > 0:
                    # The IDs are not sequential. Can't use this technique.
                    pass
                else:
                    nodes = self.tree.xpath('//case[%s=%s]' % (attribute, ids[0]))

            if len(nodes) > 1 and case_name is not None:
                # Disambiguate the possible case nodes to find the best one.
                # Initial strings take the form:
                #    2:16-cr-01152-JZB USA v. Abuarar (closed 01/26/2017)
                # Attempt to pull out just the case name.
                title_attrs = [node.xpath('./@title')[0] for node in nodes]
                strs = [title.split(' ', 1)[1].split('(closed', 1)[0] for
                        title in title_attrs]
                case_index = get_closest_match_index(case_name, strs)
                if case_index is None:
                    logger.warn("Got %s candidates, but unable to find good "
                                "match to '%s'" % (len(nodes), case_name))
                else:
                    nodes = [nodes[case_index]]

            if len(nodes) > 1:
                raise ParsingException(
                    "Multiple results returned but no way to choose the right "
                    "match. Try the case_name, office_number, or "
                    "docket_number_letters parameters?"
                )

        if nodes:
            node = nodes[0]
            return {
                u'docket_number': force_unicode(node.xpath('./@number')[0]),
                u'pacer_case_id': force_unicode(node.xpath('./@id')[0]),
                # This could be further post processed to pull out the date closed
                # and a cleaner title.
                u'title': force_unicode(node.xpath('./@title')[0]),
            }
        else:
            raise ParsingException("Unable to identify case.")


class ShowCaseDocApi(BaseReport):
    """Lookup a pacer_doc_id using the pacer_case_id and document_number

    There is a URL that's available at:

        https://ecf.nyed.uscourts.gov/cgi-bin/show_case_doc?90,406590,,

    That takes a document number and case number, and performs a redirect to a
    doc1 URL. Thus, using this API you can get the pacer_doc_id value for an
    item using only the document_number, pacer_case_id, and (optionally) the
    attachment_number.
    """
    PATH = 'cgi-bin/show_case_doc'

    def __init__(self, court_id, pacer_session=None):
        assert not court_id.endswith('b'), \
            "This API is not available at bankruptcy courts."
        super(ShowCaseDocApi, self).__init__(court_id, pacer_session)

    def query(self, pacer_case_id, document_number, attachment_number=''):
        """Query the show_case_doc endpoint and return the normalized doc1
        number.

        :param pacer_case_id: The internal PACER case ID for the case.
        :param document_number: The document number on the docket for the item.
        :param attachment_number: The attachment number of the item on the
        attachment page.
        """
        assert self.session is not None, \
            "session attribute of ShowCaseDocApi cannot be None."
        url = ('{url}?'
               '{document_number},'
               '{pacer_case_id},'
               '{attachment_number},,'.format(
                    url=self.url,
                    document_number=document_number,
                    pacer_case_id=pacer_case_id,
                    attachment_number=attachment_number,
               ))
        logger.info(u'Querying the show_doc_url endpoint with URL: %s' % url)
        # Only do a head request, else we get text content we don't need.
        self.response = self.session.head(url, allow_redirects=True)
        self.parse()

    def _parse_text(self, text):
        """Just a stub. No parsing needed here."""
        pass

    @property
    def data(self):
        """Get the URL out of the response object."""
        url = self.response.url
        if 'doc1' in url:
            return get_pacer_doc_id_from_doc1_url(url)
        else:
            raise ParsingException("Unable to get doc1-style URL. Instead got %s" % url)



from juriscraper.AbstractSite import AbstractSite


class OralArgumentSite(AbstractSite):
    """Contains generic methods for scraping data. Should be extended by all
    scrapers.

    Should not contain lists that can't be sorted by the _date_sort function.
    """

    def __init__(self, *args, **kwargs):
        super(OralArgumentSite, self).__init__(*args, **kwargs)

        self._opt_attrs = ['docket_numbers', 'judges', 'case_name_shorts',]
        self._req_attrs = ['case_dates', 'case_names', 'download_urls',
                           'blocked_statuses',]
        # For date sorting to work, case_dates must be the first item in _all_attrs.
        self._all_attrs = self._req_attrs + self._opt_attrs

        # Set all metadata to None
        for attr in self._all_attrs:
            self.__setattr__(attr, None)

    def _get_download_urls(self):
        raise NotImplementedError('`_get_download_urls()` must be implemented.')

    def _get_case_dates(self):
        raise NotImplementedError('`_get_case_dates()` must be implemented.')

    def _get_case_names(self):
        raise NotImplementedError('`_get_case_names()` must be implemented.')

    def _get_docket_numbers(self):
        return None

    def _get_judges(self):
        return None



#  Scraper for Georgia Supreme Court
# CourtID: ga
# Court Short Name: ga
# History:
#  - 2014-07-25: Created by Andrei Chelaru, reviewed by mlr
#  - 2015-07-30: MLR: Added more lenient dates.
#  - 2016-09-02: arderyp: fixed to handle slight site redesign, while accomodating legacy

import re
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.gasupreme.us/opinions/{year}-opinions/'.format(year=date.today().year)
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self.extract_cases_from_html(html)
        return html

    def extract_cases_from_html(self, html):
        paths = '//p/strong | //p/b | //p/font/strong | //p/font/b'
        for date_element in html.xpath(paths):
            string = date_element.xpath('./text()')
            try:
                string = string[0]
                # handle examples where time but no date (ga_example_3.html)
                if ':' in string and ('AM' in string or 'PM' in string):
                    continue
                # handle legacy example (ga_example.html)
                string = string.split('SUMMARIES')[0]
                date_string = re.sub(r'\W+', ' ', string)
                # handle legacy example (ga_example.html)
                if len(date_string.split()) != 3:
                    continue
                case_date = convert_date_string(date_string)
            except:
                continue
            parent = date_element.xpath('./..')[0]
            # handle legacy example (ga_example.html)
            while parent.tag != 'p':
                parent = parent.xpath('./..')[0]
            for item in parent.getnext().xpath('./li'):
                text = item.text_content()
                if text:
                    split = text.split('.', 1)
                    self.cases.append({
                        'date': case_date,
                        'url': item.xpath('//a[1]/@href')[0],
                        'docket': split[0].rstrip('.'),
                        'name': titlecase(split[1]),
                    })

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)



"""Scraper for the Texas Attorney General
CourtID: texag
Court Short Name: Texas Attorney General
"""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import convert_date_string

## WARNING: THIS SCRAPER IS FAILING:
## The page on which the hosting court was
## previously publishing opionions publically
## not shows a messgae that opinions are only
## available upon request. Unless we can convince
## the court to start publishing opinions publicly
## agian, we can get rid of this scraper.


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.target_index = 2
        self.url_path = False
        self.opinion_path = False
        self.section_path = False
        self.year_sub_path = False
        self.opinion_sub_path = False
        self.domain = 'https://texasattorneygeneral.gov'
        self.url = '%s/opinion/index-to-opinions' % self.domain
        self.back_scrape_iterable = range(2, 16)  # Hard coded for initial run
        self.select_sub_path = './/select/option[position()>1]'
        self.flat_list_path = '//a[contains(./text(), "See a flat listing of all opinions")]'
        self.target_sub_page_path_base = '//table/tbody/tr[%d]/td[2]//a/@href'
        self.target_sub_page_path = self.target_sub_page_path_base % self.target_index

    def _download(self, request_dict={}):
        """Follow top-most opinions urls on landing page to resource page"""
        # Process landing page
        landing_html = super(Site, self)._download(request_dict)
        if self.test_mode_enabled():
            # Example file should be direct resource page
            return landing_html
        # Load resource page
        url = landing_html.xpath(self.target_sub_page_path)[0]
        resource_page_html = self._get_html_tree_by_url(url, request_dict)
        flat_list_link = resource_page_html.xpath(self.flat_list_path)
        if not flat_list_link:
            return resource_page_html
        # Load flat list page for older pages with bad js
        url = flat_list_link[0].xpath('./@href')[0]
        return self._get_html_tree_by_url(url)

    def _get_case_dates(self):
        """All we have are years, so estimate middle most day of year"""
        self.set_dynamic_resource_paths()
        dates = []
        for section in self.html.xpath(self.section_path):
            year = section.xpath(self.year_sub_path)[0].text_content().strip()
            date = convert_date_string('July 2, %s' % year)
            count = len(section.xpath(self.opinion_sub_path))
            dates.extend([date] * count)
        return dates

    def _get_case_names(self):
        """No case names available"""
        return ["Untitled Texas Attorney General Opinion"] * len(self.case_dates)

    def _get_download_urls(self):
        # Some listings provide direct links, others are relative
        return [self.domain + v if self.domain not in v else v
                for v in self.html.xpath(self.url_path)]

    def _get_docket_numbers(self):
        return [option.text_content().strip() for option in self.html.xpath(self.opinion_path)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_dates)

    def _get_judges(self):
        breadcrumb = self.html.xpath('//div[contains(@class, "breadcrumb")]//li')[-1].text_content().strip()
        return [breadcrumb.split('Opinions')[0]] * len(self.case_dates)

    def _get_date_filed_is_approximate(self):
        return [True] * len(self.case_dates)

    def _download_backwards(self, index):
        self.target_index = index
        self.target_sub_page_path = self.target_sub_page_path_base % index
        self.html = self._download()

    # Across the whole history of the opinions, the court
    # has used various slightly different page html formats
    # The functions below are used to detect which format
    # the page is using, and set the path variables accordingly.

    def set_dynamic_resource_paths(self):
        self.opinion_path = self.return_opinion_path()
        self.opinion_sub_path = '.%s' % self.opinion_path
        self.url_path = self.return_url_path()
        self.section_path = self.return_section_path()
        self.year_sub_path = self.return_year_sub_path()

    def return_section_path(self):
        paths = [
            '//div[contains(@class, "panel-default")]',
            '//td[contains(p/@class, "center")]',
            '//td[contains(p/@align, "center")]',
            '//td[contains(h2/@class, "center")]',
            '//div[contains(h3/@class, "center")]',
            '//div[contains(h3/@align, "center")]',
        ]
        for path in paths:
            if self.html.xpath(path):
                return path
        raise InsanityException('No recognized path to opinion sections')

    def return_year_sub_path(self):
        parent = self.html.xpath(self.section_path)[0]
        paths = [
            './div[contains(@class, "panel-heading")]/label',
            './p[contains(@class, "center")]/strong',
            './p[contains(@align, "center")]/font/b',
            './h2[contains(@class, "center")]',
            './h3[contains(@class, "center")]',
            './h3[contains(@align, "center")]',
        ]
        for path in paths:
            if parent.xpath(path):
                return path
        raise InsanityException('No recognized path to year string')

    def return_opinion_path(self):
        paths = [
            '//select/option[contains(@value, ".pdf")]',
            '//ul/li/a[contains(@href, ".pdf")]',
        ]
        for path in paths:
            if self.html.xpath(path):
                return path
        raise InsanityException('No recognized path to opinion listings')

    def return_url_path(self):
        if '/option' in self.opinion_path:
            return '%s/@value' % self.opinion_path
        elif '/li/a' in self.opinion_path:
            return '%s/@href' % self.opinion_path
        raise InsanityException('No recognized path to url')



# Scraper for Oklahoma Court of Civil Appeals
#CourtID: oklacivapp
#Court Short Name: OK
#Author: Andrei Chelaru
#Reviewer: mlr
#Date: 2014-07-05

from datetime import date

from juriscraper.opinions.united_states.state import okla


class Site(okla.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        d = date.today()
        self.url = 'http://www.oscn.net/applications/oscn/Index.asp?ftdb=STOKCSCV&year={year}&level=1'.format(
            year=d.year
        )



# coding=utf-8
import re

from .docket_report import DocketReport
from .utils import get_nonce_from_form, get_pacer_doc_id_from_doc1_url, \
    get_pacer_seq_no_from_doc1_anchor
from ..lib.judge_parsers import normalize_judge_string
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, convert_date_string, \
    force_unicode, harmonize
from ..lib.utils import clean_court_object, previous_and_next

logger = make_default_logger()

date_regex = r'[—\d\-–/]*'


class DocketHistoryReport(DocketReport):
    assigned_to_regex = re.compile(r'(.*),\s+presiding', flags=re.IGNORECASE)
    referred_to_regex = re.compile(r'(.*),\s+referral', flags=re.IGNORECASE)
    date_filed_regex = re.compile(r'[fF]iled:\s+(%s)' % date_regex)
    date_last_filing_regex = re.compile(r'last\s+filing:\s+(%s)' % date_regex,
                                        flags=re.IGNORECASE)
    date_filed_and_entered_regex = re.compile(r'& Entered:\s+(%s)' % date_regex)

    PATH = 'cgi-bin/HistDocQry.pl'

    @property
    def data(self):
        """Get all the data back from this endpoint."""
        if self.is_valid is False:
            return {}
        data = self.metadata.copy()
        data[u'docket_entries'] = self.docket_entries
        return data

    @property
    def metadata(self):
        if self._metadata is not None:
            return self._metadata

        self._set_metadata_values()
        data = {
            u'court_id': self.court_id,
            u'docket_number': self._get_docket_number(),
            u'case_name': self._get_case_name(),
            u'date_filed': self._get_value(self.date_filed_regex,
                                           self.metadata_values,
                                           cast_to_date=True),
            u'date_last_filing': self._get_value(self.date_last_filing_regex,
                                                 self.metadata_values,
                                                 cast_to_date=True),
            u'date_terminated': self._get_value(self.date_terminated_regex,
                                                self.metadata_values,
                                                cast_to_date=True),
            u'date_discharged': self._get_value(self.date_discharged_regex,
                                                self.metadata_values,
                                                cast_to_date=True),
            u'assigned_to_str': self._get_assigned_judge(),
            u'referred_to_str': self._get_judge(self.referred_to_regex),
        }

        data = clean_court_object(data)
        self._metadata = data
        return data

    def query(self, pacer_case_id, query_type="History", order_by='asc',
              show_de_descriptions=False):
        """Query the docket history report and return the results. Because of
        the way this works, you have to hit PACER twice. Once to get a nonce,
        and a second time to make your query.

        :param pacer_case_id: The internal PACER case ID for a case.
        :param query_type: The type of query placed. Either "History" or
        "Documents".
        :param show_de_descriptions: Whether to show docket entry descriptions
        in the report.
        :param order_by: The ordering desired for the results, either 'asc' or
        'desc'.
        :return: request response object
        """
        # Set up and sanity tests
        assert self.session is not None, \
            "session attribute of DocketHistoryReport cannot be None."

        if query_type not in [u'History', u'Documents']:
            raise ValueError(u"Invalid value for 'query_type' parameter.")
        if show_de_descriptions is not True and show_de_descriptions is not False:
            raise ValueError(u"")
        if order_by not in ['asc', 'desc']:
            raise ValueError(u"Invalid value for 'order_by' parameter.")

        logger.info(u'Getting nonce for docket history report with '
                    u'pacer_case_id: %s' % pacer_case_id)
        r = self.session.get('%s?%s' % (self.url, pacer_case_id))
        nonce = get_nonce_from_form(r)

        query_params = {
            u'QueryType': query_type,
            u'sort1': order_by,
        }
        if show_de_descriptions:
            query_params['DisplayDktText'] = u'DisplayDktText'

        logger.info(u"Querying docket history report for case ID '%s' with "
                    u"params %s and nonce %s" % (pacer_case_id, query_params,
                                                 nonce))

        self.response = self.session.post(self.url + '?' + nonce,
                                          data=query_params)
        self.parse()

    @property
    def docket_entries(self):
        if self._docket_entries is not None:
            return self._docket_entries

        docket_header = './/th/text()[contains(., "Description")]'
        docket_entry_rows = self.tree.xpath(
            '//table[%s]//tr' % docket_header
        )[1:]  # Skip first row

        docket_entries = []
        for row in docket_entry_rows:
            cells = row.xpath('./td')
            if len(cells) == 3:
                # Normal row, parse the document_number, date, etc.
                de = {}
                de[u'document_number'] = clean_string(cells[0].text_content())
                if de[u'document_number'] == '':
                    de[u'document_number'] = None
                anchors = cells[0].xpath('.//a')
                if len(anchors) == 1:
                    doc1_url = anchors[0].xpath('./@href')[0]
                    de[u'pacer_doc_id'] = get_pacer_doc_id_from_doc1_url(
                        doc1_url)
                    de[u'pacer_seq_no'] = get_pacer_seq_no_from_doc1_anchor(
                        anchors[0])
                else:
                    # Unlinked minute entry; may or may not be numbered
                    de[u'pacer_doc_id'] = None
                    de[u'pacer_seq_no'] = None
                de[u'date_filed'] = self._get_date_filed(cells[1])
                de[u'short_description'] = force_unicode(cells[2].text_content())
                de[u'description'] = u''
                docket_entries.append(de)
            elif len(cells) == 1:
                # Document long description. Get it, and add it to previous de.
                desc = force_unicode(cells[0].text_content())
                label = 'Docket Text: '
                if desc.startswith(label):
                    desc = desc[len(label):]
                docket_entries[-1]['description'] = desc

        # Some docket history entries show the word "doc" instead of an entry
        # number. These items aren't on the docket itself, and so for now we
        # just skip them.
        docket_entries = [de for de in docket_entries if
                          de['document_number'] is None or
                          de['document_number'].isdigit()]
        docket_entries = clean_court_object(docket_entries)
        self._docket_entries = docket_entries
        return docket_entries

    def _get_date_filed(self, cell):
        s = clean_string(cell.text_content())
        regexes = [self.date_filed_regex, self.date_filed_and_entered_regex]
        for regex in regexes:
            m = regex.search(s)
            if m:
                return convert_date_string(m.group(1))

    def _set_metadata_values(self):
        text_nodes = self.tree.xpath('//center[not(.//table)]//text()')
        values = []
        for s in text_nodes:
            s = clean_string(force_unicode(s))
            if s:
                values.append(s)
        values.append(' '.join(values))
        self.metadata_values = values

    def _get_case_name(self):
        if self.is_bankruptcy:
            # Uses both b/c sometimes the bankr. cases have a dist-style docket
            # number.
            regexes = [self.docket_number_dist_regex,
                       self.docket_number_bankr_regex]
        else:
            regexes = [self.docket_number_dist_regex]

        matches = set()
        # Skip the last value, it's a concat of all previous values and
        # isn't needed for case name matching.
        for prev, v, nxt in previous_and_next(self.metadata_values[:-1]):
            if prev is None:
                continue
            for regex in regexes:
                match = regex.search(prev)
                if match:
                    if self.is_bankruptcy:
                        return harmonize(v)
                    for cn_regex in self.case_name_regexes:
                        cn_match = cn_regex.match(v)
                        if cn_match:
                            matches.add(cn_match.group(1))
        if len(matches) == 1:
            case_name = list(matches)[0]
        else:
            case_name = u"Unknown Case Title"
        return harmonize(case_name)

    def _get_docket_number(self):
        if self.is_bankruptcy:
            # Uses both b/c sometimes the bankr. cases have a dist-style docket
            # number.
            regexes = [self.docket_number_dist_regex,
                       self.docket_number_bankr_regex]
        else:
            regexes = [self.docket_number_dist_regex]
        nodes = self.tree.xpath('//center//font[@size="+1"]')
        string_nodes = [s.text_content() for s in nodes]
        for regex in regexes:
            for s in string_nodes:
                match = regex.search(s)
                if match:
                    return match.group(1)

    def _get_assigned_judge(self):
        if self.is_bankruptcy:
            # Look for string like "Judge: Michael J. Fox"
            for prev, v, nxt in previous_and_next(self.metadata_values[:-1]):
                if prev is not None and re.search('Judge:', prev, flags=re.I):
                    return normalize_judge_string(v)[0]
        else:
            # Look for string like "Michael J. Fox, presiding"
            return self._get_judge(self.assigned_to_regex)
